@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/main.c,v 1.83 2010/08/22 08:24:01 gleu Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/main.c,v 1.85.2.9 2011/05/25 17:13:16 gleu Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2010	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -120,6 +120,8 @@ static void stop_me(void);
 
 static int trigger_failover_command(int node, const char *command_line);
 
+static int find_primary_node(void);
+
 static struct sockaddr_un un_addr;		/* unix domain socket path */
 static struct sockaddr_un pcp_un_addr;  /* unix domain socket path for PCP */
 
@@ -148,8 +150,6 @@ static int switching = 0;		/* non 0 if I'm fail overing or degenerating */
 static int clear_cache = 0;		/* non 0 if clear chache option (-c) is given */
 static int not_detach = 0;		/* non 0 if non detach option (-n) is given */
 
-long int weight_master;	/* normalized weight of master (0-RAND_MAX range) */
-
 static int stop_sig = SIGTERM;	/* stopping signal default value */
 
 static volatile sig_atomic_t health_check_timer_expired;		/* non 0 if health check timer expired */
@@ -311,25 +311,12 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	/*
-	 * Locate pool_passwd
-	 */
-	if (strcmp("", pool_config->pool_passwd))
-	{
-		char pool_passwd[POOLMAXPATHLEN+1];
-		char dirnamebuf[POOLMAXPATHLEN+1];
-
-		strncpy(dirnamebuf, conf_file, sizeof(dirnamebuf));
-		dirname(dirnamebuf);
-		snprintf(pool_passwd, sizeof(pool_passwd), "%s/%s",
-				 dirnamebuf, pool_config->pool_passwd);
-		pool_init_pool_passwd(pool_passwd);
-	}
 
 	/*
 	 * Override debug level
 	 */
-	pool_config->debug_level = debug_level;
+	if (pool_config->debug_level == 0)
+		pool_config->debug_level = debug_level;
 
 	if (pool_config->enable_pool_hba)
 		load_hba(hba_file);
@@ -406,6 +393,22 @@ int main(int argc, char **argv)
 		write_pid_file();
 	else
 		daemonize();
+	
+	/*
+	 * Locate pool_passwd
+	 */
+	if (strcmp("", pool_config->pool_passwd))
+	{
+		char pool_passwd[POOLMAXPATHLEN+1];
+		char dirnamebuf[POOLMAXPATHLEN+1];
+		char *dirp;
+
+		strncpy(dirnamebuf, conf_file, sizeof(dirnamebuf));
+		dirp = dirname(dirnamebuf);
+		snprintf(pool_passwd, sizeof(pool_passwd), "%s/%s",
+				 dirp, pool_config->pool_passwd);
+		pool_init_pool_passwd(pool_passwd);
+	}
 
 	if (pool_semaphore_create(MAX_NUM_SEMAPHORES))
 	{
@@ -549,6 +552,9 @@ int main(int argc, char **argv)
 	retrycnt = 0;		/* reset health check retry counter */
 	sys_retrycnt = 0;	/* reset SystemDB health check retry counter */
 
+	/* Save primary node id */
+	Req_info->primary_node_id = find_primary_node();
+
 	/*
 	 * This is the main loop
 	 */
@@ -902,7 +908,10 @@ static int read_status_file(bool discard_status)
 	for (i=0;i< pool_config->backend_desc->num_backends;i++)
 	{
 		if (backend_rec.status[i] == CON_DOWN)
+		{
 			BACKEND_INFO(i).backend_status = CON_DOWN;
+			pool_log("read_status_file: %d th backend is set to down status", i);
+		}
 		else
 		{
 			BACKEND_INFO(i).backend_status = CON_CONNECT_WAIT;
@@ -1608,6 +1617,9 @@ static void failover(void)
 				 BACKEND_INFO(node_id).backend_port);
 	}
 
+	/* Save primary node id */
+	Req_info->primary_node_id = find_primary_node();
+
 	switching = 0;
 
 	/* kick wakeup_handler in pcp_child to notice that
@@ -1936,7 +1948,7 @@ static void reaper(void)
 BackendInfo *
 pool_get_node_info(int node_number)
 {
-	if (node_number >= NUM_BACKENDS)
+	if (node_number < 0 || node_number >= NUM_BACKENDS)
 		return NULL;
 
 	return &BACKEND_INFO(node_number);
@@ -2180,27 +2192,31 @@ static int trigger_failover_command(int node, const char *command_line)
 				char val = *(command_line + 1);
 				switch (val)
 				{
-					case 'p': /* port */
+					case 'p': /* failed node port */
 						snprintf(port_buf, sizeof(port_buf), "%d", info->backend_port);
 						string_append_char(exec_cmd, port_buf);
 						break;
 
-					case 'D': /* database directory */
+					case 'D': /* failed node database directory */
 						string_append_char(exec_cmd, info->backend_data_directory);
 						break;
 
-					case 'd': /* node id */
+					case 'd': /* failed node id */
 						snprintf(port_buf, sizeof(port_buf), "%d", node);
 						string_append_char(exec_cmd, port_buf);
 						break;
 
-					case 'h': /* host name */
+					case 'h': /* failed host name */
 						string_append_char(exec_cmd, info->backend_hostname);
 						break;
 
 					case 'H': /* new master host name */
 						newmaster = pool_get_node_info(get_next_master_node());
-						string_append_char(exec_cmd, newmaster->backend_hostname);
+						if (newmaster)
+							string_append_char(exec_cmd, newmaster->backend_hostname);
+						else
+							/* no vaid new master */
+							string_append_char(exec_cmd, "");
 						break;
 
 					case 'm': /* new master node id */
@@ -2213,6 +2229,11 @@ static int trigger_failover_command(int node, const char *command_line)
 						string_append_char(exec_cmd, port_buf);
 						break;
 
+					case 'P': /* old primary node id */
+						snprintf(port_buf, sizeof(port_buf), "%d", PRIMARY_NODE_ID);
+						string_append_char(exec_cmd, port_buf);
+						break;
+
 					case '%': /* escape */
 						string_append_char(exec_cmd, "%");
 						break;
@@ -2240,3 +2261,117 @@ static int trigger_failover_command(int node, const char *command_line)
 
 	return r;
 }
+
+/*
+ * Find the primary node (i.e. not standby node) and returns its node
+ * id. If no primary node is found, returns -1.
+ */
+static int find_primary_node(void)
+{
+	BackendInfo *bkinfo;
+	POOL_CONNECTION_POOL_SLOT *s;
+	POOL_CONNECTION *con; 
+	POOL_STATUS status;
+	POOL_SELECT_RESULT *res;
+	bool is_standby;
+	int i;
+
+	/* Streaming replication mode? */
+	if (pool_config->master_slave_mode == 0 ||
+		strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP))
+	{
+		/* No point to look for primary node if not in streaming
+		 * replication mode.
+		 */
+		pool_debug("find_primary_node: not in streaming replication mode");
+		return -1;
+	}
+
+	for(i=0;i<NUM_BACKENDS;i++)
+	{
+		if (!VALID_BACKEND(i))
+			continue;
+
+		/*
+		 * Check to see if this is a standby node or not.
+		 */
+		is_standby = false;
+
+		bkinfo = pool_get_node_info(i);
+		s = make_persistent_db_connection(bkinfo->backend_hostname, 
+										  bkinfo->backend_port,
+										  "postgres",
+										  pool_config->health_check_user,
+										  "");
+		if (!s)
+		{
+			pool_error("find_primary_node: make_persistent_connection failed");
+			break;
+		}
+		con = s->con;
+
+		status = do_query(con, "SELECT count(*) FROM pg_catalog.pg_proc AS p WHERE p.proname = 'pgpool_walrecrunning'",
+						  &res, PROTO_MAJOR_V3);
+		if (res->numrows <= 0)
+		{
+			pool_log("find_primary_node: do_query returns no rows");
+		}
+		if (res->data[0] == NULL)
+		{
+			pool_log("find_primary_node: do_query returns no data");
+		}
+		if (res->nullflags[0] == -1)
+		{
+			pool_log("find_primary_node: do_query returns NULL");
+		}
+		if (res->data[0] && !strcmp(res->data[0], "0"))
+		{
+			pool_log("find_primary_node: pgpool_walrecrunning does not exist");
+			free_select_result(res);
+			discard_persistent_db_connection(s);
+			return -1;
+		}
+
+		status = do_query(con, "SELECT pg_is_in_recovery() AND pgpool_walrecrunning()",
+						  &res, PROTO_MAJOR_V3);
+		if (res->numrows <= 0)
+		{
+			pool_log("find_primary_node: do_query returns no rows");
+		}
+		if (res->data[0] == NULL)
+		{
+			pool_log("find_primary_node: do_query returns no data");
+		}
+		if (res->nullflags[0] == -1)
+		{
+			pool_log("find_primary_node: do_query returns NULL");
+		}
+		if (res->data[0] && !strcmp(res->data[0], "t"))
+		{
+			is_standby = true;
+		}   
+		free_select_result(res);
+		discard_persistent_db_connection(s);
+
+		/*
+		 * If this is a standby, we continue to look for primary node.
+		 */
+		if (is_standby)
+		{
+			pool_log("find_primary_node: %d node is standby", i);
+		}
+		else
+		{
+			break;
+		}
+	}
+
+	if (i == NUM_BACKENDS)
+	{
+		pool_log("find_primary_node: no primary node found");
+		return -1;
+	}
+
+	pool_log("find_primary_node: primary node id is %d", i);
+	return i;
+}
