@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pcp_child.c,v 1.20 2010/08/05 23:37:43 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pcp_child.c,v 1.20.2.1 2011/02/23 13:19:32 kitagawa Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
@@ -779,7 +779,7 @@ pcp_do_child(int unix_fd, int inet_fd, char *pcp_conf_file)
 					   !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP))) ||
 					(MASTER_SLAVE &&
 					 !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
-					 node_id == REAL_MASTER_NODE_ID))
+					 node_id == PRIMARY_NODE_ID))
 				{
 					int len;
 					char *msg;
