@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.89.2.7 2011/02/23 13:19:32 kitagawa Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_proto_modules.c,v 1.89.2.9 2011/05/26 05:04:10 kitagawa Exp $
  * 
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -700,6 +700,8 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 				msg->len = len;
 				msg->contents = contents;
+
+				query_context->rewritten_query = rewrite_query;
 			}
 		}
 	}
@@ -1424,6 +1426,7 @@ POOL_STATUS BindComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backen
 POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
 	POOL_SESSION_CONTEXT *session_context;
+	POOL_STATUS status;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -1433,6 +1436,10 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 		return POOL_END;
 	}
 
+	/* Send CloseComplete(3) to frontend before removing the target message */
+	status = SimpleForwardToFrontend('3', frontend, backend);
+
+	/* Remove the target message */
 	if (session_context->uncompleted_message)
 	{
 		pool_remove_sent_message(session_context->uncompleted_message->kind,
@@ -1445,7 +1452,7 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 		return POOL_END;
 	}
 
-	return SimpleForwardToFrontend('3', frontend, backend);
+	return status;
 }
 
 POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
