@@ -1,12 +1,12 @@
 /* -*-pgsql-c-*- */
 /*
  *
- * $Header: /cvsroot/pgpool/pgpool-II/pool_query_context.c,v 1.31 2010/09/01 06:10:08 kitagawa Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_query_context.c,v 1.32.2.7 2011/05/26 05:04:09 kitagawa Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2010	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -239,6 +239,9 @@ bool pool_is_node_to_be_sent_in_current_query(int node_id)
 {
 	POOL_SESSION_CONTEXT *sc;
 
+	if (RAW_MODE)
+		return node_id == REAL_MASTER_NODE_ID;
+
 	sc = pool_get_session_context();
 	if (!sc)
 		return true;
@@ -298,7 +301,16 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	 */
 	if (!strncasecmp(query, NO_LOAD_BALANCE, NO_LOAD_BALANCE_COMMENT_SZ))
 	{
-		pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+		pool_set_node_to_be_sent(query_context,
+								 MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID);
+		for (i=0;i<NUM_BACKENDS;i++)
+		{
+			if (query_context->where_to_send[i])
+			{
+				query_context->virtual_master_node_id = i;
+				break;
+			}
+		}
 		return;
 	}
 
@@ -323,7 +335,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 		/* Should be sent to primary only? */
 		if (dest == POOL_PRIMARY)
 		{
-			pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+			pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 		}
 		/* Should be sent to both primary and standby? */
 		else if (dest == POOL_BOTH)
@@ -346,7 +358,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 				 *	transaction isolation level is not SERIALIZABLE)
 				 * we might be able to load balance.
 				 */
-				if (TSTATE(backend, MASTER_NODE_ID) == 'I' ||
+				if (TSTATE(backend, PRIMARY_NODE_ID) == 'I' ||
 					(!pool_is_writing_transaction() &&
 					 !pool_is_failed_transaction() &&
 					 pool_get_transaction_isolation() != POOL_SERIALIZABLE))
@@ -364,7 +376,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 						pool_config->delay_threshold &&
 						bkinfo->standby_delay > pool_config->delay_threshold)
 					{
-						pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
 
 					/*
@@ -373,7 +385,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 					 */
 					else if (pool_has_function_call(node))
 					{
-						pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
 
 					/*
@@ -390,7 +402,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 					 */
 					else if (pool_has_system_catalog(node))
 					{
-						pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
 
 					/*
@@ -399,7 +411,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 					 */
 					else if (pool_has_temp_table(node))
 					{
-						pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+						pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 					}
 
 					else
@@ -411,49 +423,15 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 				else
 				{
 					/* Send to the primary only */
-					pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+					pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 				}
 			}
 			else
 			{
 				/* Send to the primary only */
-				pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
+				pool_set_node_to_be_sent(query_context, PRIMARY_NODE_ID);
 			}
 		}
-
-		/* PREPARE? */
-		if (IsA(node, PrepareStmt))
-		{
-			/* Make sure that same prepared statement does not exist */
-			if (pool_get_prep_where(((PrepareStmt *)node)->name) == NULL)
-			{
-				/* Save the send map */
-				pool_add_prep_where(((PrepareStmt *)node)->name, query_context->where_to_send);
-			}
-		}
-
-		/*
-		 * EXECUTE?
-		 */
-		else if (IsA(node, ExecuteStmt))
-		{
-			bool *wts;
-
-			wts = pool_get_prep_where(((ExecuteStmt *)node)->name);
-			if (wts)
-			{
-				/* Inherit same map from PREPARE */
-				pool_copy_prep_where(wts, query_context->where_to_send);
-			}
-		}
-
-		/*
-		 * DEALLOCATE?
-		 */
-		else if (IsA(node, DeallocateStmt))
-		{
-			where_to_send_deallocate(query_context, node);
-		}
 	}
 	else if (REPLICATION || PARALLEL_MODE)
 	{
@@ -481,13 +459,6 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 				pool_set_node_to_be_sent(query_context, REAL_MASTER_NODE_ID);
 			}
 		}
-		/*
-		 * DEALLOCATE?
-		 */
-		else if (IsA(node, DeallocateStmt))
-		{
-			where_to_send_deallocate(query_context, node);
-		}
 		else
 		{
 			/* send to all nodes */
@@ -500,6 +471,29 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 		return;
 	}
 
+	/*
+	 * EXECUTE?
+	 */
+	if (IsA(node, ExecuteStmt))
+	{
+		POOL_SENT_MESSAGE *msg;
+
+		msg = pool_get_sent_message('Q', ((ExecuteStmt *)node)->name);
+		if (!msg)
+			msg = pool_get_sent_message('P', ((ExecuteStmt *)node)->name);
+		if (msg)
+			pool_copy_prep_where(msg->query_context->where_to_send,
+								 query_context->where_to_send);
+	}
+
+	/*
+	 * DEALLOCATE?
+	 */
+	else if (IsA(node, DeallocateStmt))
+	{
+		where_to_send_deallocate(query_context, node);
+	}
+
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
 		if (query_context->where_to_send[i])
@@ -508,10 +502,10 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 			break;
 		}
 	}
+
 	return;
 }
 
-
 /*
  * Send query and wait for response
  * string:
@@ -572,15 +566,41 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context, char *string,
 			continue;
 		}
 
-		per_node_statement_log(backend, i, string);
-
 		if (*kind == '\0')
 		{
+			per_node_statement_log(backend, i, string);
+
 			if (send_simplequery_message(CONNECTION(backend, i), len, string, MAJOR(backend)) != POOL_CONTINUE)
 				return POOL_END;
 		}			
 		else
 		{
+			if (pool_config->log_per_node_statement)
+			{
+				char msgbuf[QUERY_STRING_BUFFER_LEN];
+
+				if (*kind == 'P' || *kind == 'E')
+				{
+					char *stmt;
+
+					if (query_context->rewritten_query)
+						stmt = query_context->rewritten_query;
+					else
+						stmt = query_context->original_query;
+
+					if (*kind == 'P')
+						snprintf(msgbuf, sizeof(msgbuf), "Parse: %s", stmt);
+					else
+						snprintf(msgbuf, sizeof(msgbuf), "Execute: %s", stmt);
+				}
+				else
+				{
+					snprintf(msgbuf, sizeof(msgbuf), "%c message", *kind);
+				}
+
+				per_node_statement_log(backend, i, msgbuf);
+			}
+
 			if (send_extended_protocol_message(backend, i, kind, len, string) != POOL_CONTINUE)
 				return POOL_END;
 		}
@@ -973,7 +993,7 @@ static
 void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 {
 	DeallocateStmt *d = (DeallocateStmt *)node;
-	bool *wts;
+	POOL_SENT_MESSAGE *msg;
 
 	/* DELLOCATE ALL? */
 	if (d->name == NULL)
@@ -983,24 +1003,16 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 	}
 	else
 	{
-		wts = pool_get_prep_where(d->name);
-		if (wts)
+		msg = pool_get_sent_message('Q', d->name);
+		if (!msg)
+			msg = pool_get_sent_message('P', d->name);
+		if (msg)
 		{
-			/* Inherit same map from PREPARE */
-			pool_copy_prep_where(wts, query_context->where_to_send);
-			return;
-		}
-		else
-		{
-			PreparedStatement *ps;
-
-			ps = pool_get_prepared_statement_by_pstmt_name(d->name);
-			if (ps && ps->qctxt)
-			{
-				pool_copy_prep_where(ps->qctxt->where_to_send, query_context->where_to_send);
-				return;
-			}
+			/* Inherit same map from PREPARE or PARSE */
+			pool_copy_prep_where(msg->query_context->where_to_send,
+								 query_context->where_to_send);
 		}
+		return;
 	}
 	/* prepared statement was not found */
 	pool_setall_node_to_be_sent(query_context);
@@ -1114,12 +1126,9 @@ bool is_2pc_transaction_query(Node *node, char *query)
 }
 
 /*
- * Set query state, if specified state less than current state
- * state:
- *  0: before parse   1: parse done     2: bind done
- *  3: describe done  4: execute done  -1: in error
+ * Set query state, if a current state is before it than the specified state.
  */
-void pool_set_query_state(POOL_QUERY_CONTEXT *query_context, short state)
+void pool_set_query_state(POOL_QUERY_CONTEXT *query_context, POOL_QUERY_STATE state)
 {
 	int i;
 
@@ -1129,16 +1138,41 @@ void pool_set_query_state(POOL_QUERY_CONTEXT *query_context, short state)
 		return;
 	}
 
-	if (state < -1 || state > 4)
-	{
-		pool_error("pool_set_query_state: invalid query state: %d", state);
-		return;
-	}
-
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
 		if (query_context->where_to_send[i] &&
-			query_context->query_state[i] < state)
+			statecmp(query_context->query_state[i], state) < 0)
 			query_context->query_state[i] = state;
 	}
 }
+
+int statecmp(POOL_QUERY_STATE s1, POOL_QUERY_STATE s2)
+{
+	int ret;
+
+	switch (s2) {
+		case POOL_UNPARSED:
+			ret = (s1 == s2) ? 0 : 1;
+			break;
+		case POOL_PARSE_COMPLETE:
+			if (s1 == POOL_UNPARSED)
+				ret = -1;
+			else
+				ret = (s1 == s2) ? 0 : 1;
+			break;
+		case POOL_BIND_COMPLETE:
+			if (s1 == POOL_UNPARSED || s1 == POOL_PARSE_COMPLETE)
+				ret = -1;
+			else
+				ret = (s1 == s2) ? 0 : 1;
+			break;
+		case POOL_EXECUTE_COMPLETE:
+			ret = (s1 == s2) ? 0 : -1;
+			break;
+		default:
+			ret = -2;
+			break;
+	}
+
+	return ret;
+}
